//=======================================================================
// This file is a component of PsyMUD written by Peter M. Brown
// copyright            : (C) 2000 by Peter M Brown
// email                : pmbrown@cablespeed.com
//=======================================================================
//                                                                          
// This program is free software; you can redistribute it and/or modify  
// it under the terms of the GNU General Public License as published by  
// the Free Software Foundation; either version 2 of the License, or     
// (at your option) any later version. Pleae keep all headers and 
// documentation (including development credits) intact.
//                                                                          
//=======================================================================



#include <errno.h>
#include <fcntl.h>
#include <unistd.h>		// for usleep()

#include "CPsyMudApp.h"
#include "CPsyMudCommand.h"
#include "CPsyMudCommandProcessor.h"
#include "mudversion.h"
#include "timeutil.h"
#include "stringutil.h"

#include "CPsyMudRoom.h"

#include <iostream>

CPsyMudApp	g_Mud ;

CPsyMudApp::CPsyMudApp()
{
	// don't log anything in here because we have no guarantee
	// that the log object has been constructed

	return ;
}

CPsyMudApp::~CPsyMudApp()
{
	// don't log anything in here because log may already be destroyed

	//g_Log.Log(CPsyMudLog::logTrace, "PsyMud Application Object destroyed.") ;

	return ;
}



void CPsyMudApp::SetPort(unsigned int iNewPort)
{
	m_uiPort = iNewPort ;
	g_Log.Log(CPsyMudLog::logGeneral, "Port set to %u", m_uiPort) ;

	return ;
}

void CPsyMudApp::SetRestricted(bool bRestricted)
{
	m_bRestricted = bRestricted ;

	if (m_bRestricted)
	{
		g_Log.Log(CPsyMudLog::logGeneral, "Restricted mode. No new players will be allowed in mud.") ;
	}
	else
	{
		g_Log.Log(CPsyMudLog::logGeneral, "Open mode. New players will be accepted in mud.") ;
	}

	return ;
}


void CPsyMudApp::SetUniverseDirectory(const char * szUniverseDirectory) throw(CPsyMudAppException)
{
	m_Universe.SetDirectory(szUniverseDirectory) ;

	g_Log.Log(CPsyMudLog::logGeneral, "Universe directory set to \"%s\"", szUniverseDirectory) ;
	
	return ;
}





void CPsyMudApp::GameLoop()
{
	g_Log.Log(CPsyMudLog::logTrace, "==================================================") ;
	g_Log.Log(CPsyMudLog::logTrace, "Entering main game loop.") ;
	g_Log.Log(CPsyMudLog::logTrace, "==================================================") ;

	bool 	bContinueRunning = true ;
	//int		iCurrentPlayerIndex = 0 ;
	//int		iNumEmptyCycles = 0 ;			// number of iterations where the playlist was empty (no players)
	int		iNumMissedPulses = 0 ;			// number of pulses we missed by sleeping too long

	struct timeval	tmStartOfProcessing ;
	struct timeval	tmEndOfProcessing ;
	struct timeval	tmTotalProcessing ;
	//struct timeval	tmAfterProcessing ;
	struct timeval	tmPauseDuration ;	
	//struct timeval	tmCurrentTime ;
	struct timeval	tmOneSinglePulse ;

	tmOneSinglePulse.tv_sec = 0 ;
	tmOneSinglePulse.tv_usec = MICROSECONDS_PER_PULSE ;

	while (bContinueRunning)
	{
		// accept new players into game
		
			// does the FD_ZERO on the descriptors, and the FD_SET on the
			// mother file descriptor
		m_comm.ResetDescriptorSets() ;


		if (m_playlist.empty())
		{
			// no connections.  go to sleep.  May need to change this if we
			// have any processing which absolutely must occur even if the player
			// list is empty. (scripts, etc)
			g_Log.Log(CPsyMudLog::logGeneral, "The active player list is empty.  Going to hibernate.") ;
			m_comm.GoToSleep() ;
	
			// we don't want to include the GoToSleep() time in the total
			// processing time, so we reset it here.  Otherwise we'd log
			// missed pulses for the whole time the mud was empty.
			gettimeofday(&tmStartOfProcessing, NULL) ;
		}
		else
		{
			// nothing to do here for now
		}

		// end of previous loop is here. "start" of loop is after PauseExecution call
		gettimeofday(&tmEndOfProcessing, NULL) ;


		// figure out how long the last loop took
		CalculateTimeDifference(tmTotalProcessing, tmEndOfProcessing, tmStartOfProcessing) ;
		//cerr << "DEBUG: Total processing time was " << tmTotalProcessing.tv_sec << ":" << tmTotalProcessing.tv_usec << endl ;

		if (tmTotalProcessing.tv_sec == 0 && tmTotalProcessing.tv_usec < MICROSECONDS_PER_PULSE)
		{
			// last loop took less than one pulse to complete
			iNumMissedPulses = 0 ;
		}
		else
		{
			// figure out how many pulses we missed
			iNumMissedPulses = tmTotalProcessing.tv_sec * PULSES_PER_SECOND ;
			iNumMissedPulses += (tmTotalProcessing.tv_usec / MICROSECONDS_PER_PULSE) ;

			// chances are that we didn't miss an exact number of pulses, so
			// we figure out the modulus here
			tmTotalProcessing.tv_sec = 0 ;
			tmTotalProcessing.tv_usec = tmTotalProcessing.tv_usec % MICROSECONDS_PER_PULSE ;
		}

		// figure out how long we need to pause
		CalculateTimeDifference(tmPauseDuration, tmOneSinglePulse, tmTotalProcessing) ;

		//cerr << "DEBUG: Pause Duration is " << tmPauseDuration.tv_sec << ":" << tmPauseDuration.tv_usec << endl ;

		// sleep.  Need to do this to keep the mud on fairly constant time
		// no matter how many people are actually logged in.  Also gives the
		// ports time to actually queue some input at the OS level
		PauseExecution(tmPauseDuration) ;

		// get start time of next loop
		gettimeofday(&tmStartOfProcessing, NULL) ;


		//non-blocking poll for any acitivity on the ports
		m_comm.PollForActivity() ;
	

		// accept new connections
		AcceptNewConnections() ;


		// TO DO: Kick Out people in exception set

		
		// TO DO: Kick out people who have the NeedsToBeDisconnected flag set in their ConnInfo object
		DisconnectUsers() ;


		// process any input
		ProcessPendingInput() ;

		// TO DO: actually execute commands that are in the input we just picked up
		InterpretInput() ;

		//increment missed pulses
		iNumMissedPulses++ ;

		// this should only happen if we miss max_int (signed) pulses.  could conceivably happen, though.
		if (iNumMissedPulses < 0)
		{
			g_Log.Log(CPsyMudLog::logError, "Missed pulses was negative!  Did we wrap? : %d", iNumMissedPulses) ;
			iNumMissedPulses = 1 ;
		}
		else if (iNumMissedPulses > 20 * PULSES_PER_SECOND)
		{
			g_Log.Log(CPsyMudLog::logError, "Missed more than 20 seconds of pulses. System may be overloaded or something is blocking. iNumMissedPulses was  : %d", iNumMissedPulses) ;
			iNumMissedPulses = 20 * PULSES_PER_SECOND ;
		}
	
		// ProcessPendingOutput() ;
		
		while (iNumMissedPulses--)
		{
			// increment pulse, rolling over when number gets large
			IncrementPulse() ;

			// perform heartbeat processing
			PerformHeartbeatProcessing() ;
		}

		ProcessPendingOutput() ;	
	}

	g_Log.Log(CPsyMudLog::logTrace, "Exiting main game loop.") ;
}

void CPsyMudApp::PauseExecution(struct timeval& tmPauseDuration)
{

	usleep(tmPauseDuration.tv_usec + 1000000 * tmPauseDuration.tv_sec) ;
}

// loops through all input in character queues
void CPsyMudApp::InterpretInput()
{
	CPsyMudCommand * cmd ;
	
	string sCommandLine ;
	string sCommand ;
	string sRestOfCommandLine ;
	

	for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
	{
		if ((*p)->GetConnectionInfo()->HasPendingCommandInInputBuffer())
		{
			//cerr << "DEBUG: Player has pending command input in buffer." << endl ;

			sCommandLine = (*p)->GetConnectionInfo()->ChunkFirstCommandFromInputBuffer() ;

			// split the command
			SplitString(sCommand, sRestOfCommandLine, sCommandLine, ' ') ;

			cerr << "DEBUG: InterpretInput : cmd[" << sCommand.c_str() << "] rest[" << sRestOfCommandLine.c_str() << "] buf[" << sCommandLine.c_str() << "]" << endl ;

			if (sCommand == "")
			{
				(*p)->GetConnectionInfo()->AddToOutputBuffer(string("")) ;		// Kludge to print prompt again
			}
			else
			{
				// first do some substitution to see if command is an alias

				string sAliasDefinition = (*p)->GetAliasDefinition(sCommand) ;

				if (sAliasDefinition != "")
				{
					// TO DO: Perform macro substituion for things like $*, $1, $2 etc.

					sCommandLine = sAliasDefinition ;

						// split the command - again
					SplitString(sCommand, sRestOfCommandLine, sCommandLine, ' ') ;
			
					cerr << "DEBUG: InterpretInput Alias expanded to : cmd[" << sCommand.c_str() << "] rest[" << sRestOfCommandLine.c_str() << "]" << endl ;
				}

				// find the command
				cmd = g_CommandProcessor.FindCommand(sCommand.c_str()) ;
	
				// actually process the command
				if (cmd)
				{
					cmd->ProcessCommand(sRestOfCommandLine.c_str(), (*p)) ;
				}
				else
				{
					static string sOops ;
					sOops = "Sorry, I'm not sure what you mean by \"" + sCommand + "\"\r\n" ;

					(*p)->GetConnectionInfo()->AddToOutputBuffer(sOops.c_str()) ;
				}
			}
		}
	}	
}

void CPsyMudApp::DisconnectUsers()
{
	string sMsg ;

	for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
	{
		if ((*p)->GetConnectionInfo()->NeedsToBeDisconnectedDueToCommError())
		{
			g_Log.Log(CPsyMudLog::logGeneral, "Player had comm error.  Forcefully logging them off.") ;
			m_comm.RemoveConnection(*(*p)->GetConnectionInfo()) ;
			m_playlist.erase(p) ;		
		}
		else if ((*p)->GetConnectionInfo()->NeedsToBeDisconnectedDueToQuit())
		{
			(*p)->GetConnectionInfo()->AddToOutputBuffer(m_Universe.GetLogoutBanner()) ;

			sMsg = (*p)->GetName() + string(" disappears in a brilliant burst of ") ;
			sMsg += COLOR_BRIGHT_YELLOW + string("mul") ;
			sMsg += COLOR_BRIGHT_WHITE + string("ti") ;
			sMsg += COLOR_BRIGHT_CYAN + string("col") ;
			sMsg += COLOR_BRIGHT_WHITE + string("or") ;
			sMsg += COLOR_BRIGHT_YELLOW + string("ed") ;
			sMsg += COLOR_NORMAL + string(" light.\r\n") ;	

			UniverseMessage(sMsg, (*p)) ;

			g_Log.Log(CPsyMudLog::logGeneral, "Player quit nicely.") ;
			

			m_comm.RemoveConnection(*(*p)->GetConnectionInfo()) ;
			m_playlist.erase(p) ;		
		}
	}
}

void CPsyMudApp::ProcessPendingInput()
{
	for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
	{
		(*p)->GetConnectionInfo()->ProcessInput() ;
	}
}

void CPsyMudApp::ProcessPendingOutput()
{
	for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
	{
		if ((*p)->GetConnectionInfo()->HasPendingOutput())
		{
			//cerr << "DEBUG: Processing pending output.  PCharacter has output pending." << endl ;

			(*p)->GetConnectionInfo()->ProcessOutput((*p)->BuildPrompt()) ;	
		}
	}
}

void CPsyMudApp::IncrementPulse()
{
	m_iCurrentPulse++ ;
	m_iCurrentPulse = m_iCurrentPulse > PULSES_BETWEEN_PULSE_ROLLOVER ? 0 : m_iCurrentPulse ;
}


void CPsyMudApp::AcceptNewConnections()
{
	CPsyMudConnInfo * inf ;
	CPsyMudPCharacter * pc ;	

	string sMsg ;

	inf = m_comm.AcceptNewConnAndReturnInfo() ;

	if (inf != NULL)
	{

		cerr << "DEBUG: Before adding to player list, NeedsToBeDisconnectedDueToQuit = " << inf->NeedsToBeDisconnectedDueToQuit() << endl ;

		//if (!m_bMudFull)
		//{
			inf->AddToOutputBuffer(m_Universe.GetLoginBanner()) ;

			pc = new CPsyMudPCharacter(inf) ;
			pc->SetState(PCHAR_STATE_NEEDS_LOGIN_PROMPT) ;

			// TEMP!!!
			pc->SetLocation(m_Universe.GetStartRoom(0)) ;
			// End Temp!!!

			m_playlist.insert(m_playlist.end(), pc) ;

			sMsg = pc->GetName() + string(" appears with a brilliant burst of ") ;
			sMsg += COLOR_BRIGHT_YELLOW + string("mul") ;
			sMsg += COLOR_BRIGHT_WHITE + string("ti") ;
			sMsg += COLOR_BRIGHT_CYAN + string("col") ;
			sMsg += COLOR_BRIGHT_WHITE + string("or") ;
			sMsg += COLOR_BRIGHT_YELLOW + string("ed") ;
			sMsg += COLOR_NORMAL + string(" light.\r\n") ;	

			UniverseMessage(sMsg, pc) ;
			
			cerr << "DEBUG: New connection accepted.  Total connections are now " << m_playlist.size() << endl ;
		//}
		//else
		//{
		//	inf->AddToOutputBuffer("Sorry, but the mud is currently full
		//}

		cerr << "DEBUG: After adding to player list, NeedsToBeDisconnectedDueToQuit = " << inf->NeedsToBeDisconnectedDueToQuit() << endl ;
		cerr << "       -- pc->NeedsToBeDisconnectedDueToQuit = " << pc->GetConnectionInfo()->NeedsToBeDisconnectedDueToQuit() << endl ;


		// TEMP!!!!
		NewCharacterSetup(pc) ;
		// END TEMP!
	}
}


void CPsyMudApp::HealCharacters()
{
	string sHealDesc ;

	for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
	{
		sHealDesc = (*p)->NaturallyHealPoints() ;

		if (sHealDesc != "")
		{
			(*p)->GetConnectionInfo()->AddToOutputBuffer(sHealDesc) ;
		}
	}
}

void CPsyMudApp::PerformHeartbeatProcessing()
{
	if (!(m_iCurrentPulse % PULSES_BETWEEN_FIGHT_UPDATES))
	{
		// perform fight updates		
	}
	
	if (!(m_iCurrentPulse % PULSES_BETWEEN_NPC_UPDATES))
	{
		// perform movement and other NPC updates

		// TEMP !!!!!!!
		for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
		{
		//	(*p)->GetConnectionInfo()->AddToOutputBuffer("[NPC Update Cycle].\r\n") ;
		}
		// end temp!
	}

	if (!(m_iCurrentPulse % PULSES_BETWEEN_TIME_UPDATES))
	{
		// update mud time
	}

	if (!(m_iCurrentPulse % PULSES_BETWEEN_WEATHER_UPDATES))
	{
		// TEMP !!!!!!!
		for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
		{
		//	(*p)->GetConnectionInfo()->AddToOutputBuffer("[Weather Update Cycle].\r\n") ;
		}
	}

	if (!(m_iCurrentPulse % PULSES_BETWEEN_POINT_UPDATES))
	{
		HealCharacters() ;
	}

	if (!(m_iCurrentPulse % PULSES_BETWEEN_SPELL_AFFECT_UPDATES))
	{
		// update spell effects
		// TEMP !!!!!!!
		for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
		{
		//	(*p)->GetConnectionInfo()->AddToOutputBuffer("[Spell Affect Update Cycle].\r\n") ;
		}
	}

	if (!(m_iCurrentPulse % PULSES_BETWEEN_CRASH_SAVES))
	{
		// save the whole thing to the universe database
		// TEMP !!!!!!!
		for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
		{
		//	(*p)->GetConnectionInfo()->AddToOutputBuffer("[Crash Save Cycle].\r\n") ;
		}
	}

	if (!(m_iCurrentPulse % PULSES_BETWEEN_USAGE_UPDATE))
	{
		// update usage logs
		// TEMP !!!!!!!
		for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
		{
		//	(*p)->GetConnectionInfo()->AddToOutputBuffer("[Usage Log Update Cycle].\r\n") ;
		}
	}
}

void CPsyMudApp::Run()
{
	//cerr << "DEBUG: In CPsyMudApp::Run()" << endl ;
	g_Log.Log(CPsyMudLog::logTrace, "PsyMud Application Run() Started.") ;
	g_Log.Log(CPsyMudLog::logGeneral, " - Version %d.%d.%d", MUDVERSIONMAJOR, MUDVERSIONMINOR, MUDVERSIONREVISION) ;
	g_Log.Log(CPsyMudLog::logGeneral, " - Built On %s", "<TO DO: insert date>") ;
	g_Log.Log(CPsyMudLog::logGeneral, " - Built By %s on %s", "<TO DO: insert userid>", "<TO DO: insert machine>") ;

	// Load Universe
	m_Universe.LoadUniverse() ;

	// Verify all object links
	// Start scripts etc.
	// Start listener

	m_comm.SetPort(m_uiPort) ;
	m_comm.InitializeSocket() ;

	// allow people to log in, run the game etc.
	GameLoop() ;

	g_Log.Log(CPsyMudLog::logTrace, "PsyMud Application Run() Stopped.") ;

	return ;
}

// Message Stuff ========================================================

void CPsyMudApp::RoomMessage(const string& sMessage, CPsyMudRoom * room, CPsyMudPCharacter * pchExclude = NULL)
{
	for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
	{
		if (((*p)->GetLocation() == room) && (pchExclude == NULL || (*p) != pchExclude))
		{
			(*p)->GetConnectionInfo()->AddToOutputBuffer(sMessage) ;
		}
	}
}
//void CPsyMudApp::ZoneMessage(const string& sMessage, CPsyMudZone * zone, CPsyMudPCharacter * pchExclude = NULL)
//{
//}

//void CPsyMudApp::WorldMessage(const string& sMessage, CPsyMudWorld * world, CPsyMudPCharacter * pchExclude = NULL)
//{
//}

void CPsyMudApp::UniverseMessage(const string& sMessage, CPsyMudPCharacter * pchExclude = NULL)
{
	for (vector<CPsyMudPCharacter *>::iterator p = m_playlist.begin() ; p < m_playlist.end() ; p++)
	{
		if (pchExclude == NULL || (*p) != pchExclude)
		{
			(*p)->GetConnectionInfo()->AddToOutputBuffer(sMessage) ;
		}
	}
}


void CPsyMudApp::NewCharacterSetup(CPsyMudPCharacter * pch)
{
	pch->AddAlias(string("n"), string("go north")) ;
	pch->AddAlias(string("s"), string("go south")) ;
	pch->AddAlias(string("e"), string("go east")) ;
	pch->AddAlias(string("w"), string("go west")) ;
	pch->AddAlias(string("u"), string("go up")) ;
	pch->AddAlias(string("d"), string("go down")) ;
	pch->AddAlias(string("ne"), string("go northeast")) ;
	pch->AddAlias(string("nw"), string("go northwest")) ;
	pch->AddAlias(string("se"), string("go southeast")) ;
	pch->AddAlias(string("sw"), string("go southwest")) ;

	pch->GetConnectionInfo()->AddToOutputBuffer(string("Directional aliases added.  Type \"alias\" to view them.\n\r")) ;

	// TEMP!!
	//pch->GetInventory()->AddItem(m_Universe.GetItemByNumber(5001)) ;
	//pch->GetConnectionInfo()->AddToOutputBuffer(string("TEMP! You have a pen added to your inventory.\n\r")) ;


}


