//=======================================================================
// This file is a component of PsyMUD written by Peter M. Brown
// copyright            : (C) 2000 by Peter M Brown
// email                : pmbrown@cablespeed.com
//=======================================================================
//                                                                          
// This program is free software; you can redistribute it and/or modify  
// it under the terms of the GNU General Public License as published by  
// the Free Software Foundation; either version 2 of the License, or     
// (at your option) any later version. Pleae keep all headers and 
// documentation (including development credits) intact.
//                                                                          
//=======================================================================



#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include <sys/time.h>

#include <algorithm>		// find and other standard algorithms

#include <errno.h>
#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>

#include "CPsyMudComm.h"
#include "CPsyMudLog.h"

#include "color.h"

CPsyMudComm::CPsyMudComm()
{
	// don't log anything in here because log may not yet be constructed

	//g_Log.Log(CPsyMudLog::logTrace, "CPsyMudComm constructor called") ;
}

CPsyMudComm::~CPsyMudComm()
{
	// don't log anything in here because log may already be destroyed

	//g_Log.Log(CPsyMudLog::logTrace, "CPsyMudComm destructor called") ;

		// empty the list
	FD_ZERO(&m_InputSet) ;
	FD_ZERO(&m_OutputSet) ;
	FD_ZERO(&m_ExceptionSet) ;

	for (vector<CPsyMudConnInfo *>::iterator p = m_lstConnections.begin() ; p < m_lstConnections.end() ; p++)
	{
		(*p)->CloseConnection() ;
	}

	m_lstConnections.clear() ;
}

void CPsyMudComm::InitializeSocket()
{
	struct sockaddr_in	srv ;
	struct hostent *	hostpointer ;
	char 				hostname[MAX_HOSTNAME+1] ;
	int 				i ;			// for setsockopt
	

	// create the socket
	if ((m_fdMother = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		// error creating socket.  write to log.  for now just cerr
		//cerr << "System Error : Error creating socket.  Error with socket() " << strerror(errno) << endl ;	
		g_Log.Log(CPsyMudLog::logError, "Error creating socket with socket().  %s", strerror(errno)) ;
	}
	else
	{
		// want to reuse the local address
		if (setsockopt(m_fdMother, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0)
		{
			// error setting socket options  write to log.  for now just cerr
			//cerr << "System Error : Error setting socket options.  Error with setsockopt() " << strerror(errno) << endl ;	
			g_Log.Log(CPsyMudLog::logError, "Error setting socket options with setsockopt().  %s", strerror(errno)) ;
		}
		
		memset(hostname, 0, MAX_HOSTNAME+1) ;

		// get our host name
		if (gethostname(hostname, MAX_HOSTNAME) < 0)
		{
			// error getting host name.  write to log.  for now just cerr
			cerr << "System Error : Error getting host name.  Error with gethostname() " << strerror(errno) << endl ;	
			g_Log.Log(CPsyMudLog::logError, "Error getting host name with gethostname().  %s", strerror(errno)) ;
		}
		else
		{
			cerr << "System Info : gethostname() Host name is " << hostname << endl ;
		}

		// get the DNS info struct about our host.  Mainly this is just to
		// ge the h_addrtype.  As of this writing, ot is always AF_INET in linux,
		// but this is considered the "proper" way to do things.
		hostpointer = gethostbyname(hostname) ;

		// this would suck.  Dunno what would make this happen.  No DNS available, I guess.
		// what would you be doing running a MUD on a machine without DNS available <g>
		if (hostpointer == NULL)
		{
			// error getting host name.  write to log.  for now just cerr
			//cerr << "System Error : Error getting host name.  Error with gethostbyname() " << strerror(errno) << endl ;	
			g_Log.Log(CPsyMudLog::logError, "Error getting host information with gethostbyname().  %s", strerror(errno)) ;
		}
		else
		{
			cerr << "System Info : gethostbyname() Host name is " << hostpointer->h_name << endl ;
		}

		// initialize and setup the server structure
		memset(&srv, 0, sizeof(srv)) ;

		srv.sin_family 	= hostpointer->h_addrtype ;
		srv.sin_port	= htons(m_iPort) ;

		//bind socket.  should probably put some retries in here or something
		if (bind(m_fdMother, (struct sockaddr*)&srv, sizeof(srv)) < 0)
		{
			// error binding socket
			cerr << "System Error : Error binding socket.  Error with bind() " << strerror(errno) << endl ;				
			g_Log.Log(CPsyMudLog::logError, "Error binding socket with bind().  %s", strerror(errno)) ;
		}

		cerr << "System Info : About to start listening on port " << m_iPort << endl ;

			// listen on socket with max of 5 queued connection attempts waiting
		listen(m_fdMother, 5) ;			
	}

	return ;
}


CPsyMudConnInfo * CPsyMudComm::AcceptNewConnAndReturnInfo()
{

	struct sockaddr_in		insockaddr ;
	size_t 					i ;
	int						newsockfd ;
	CPsyMudConnInfo *		inf ;

	if (FD_ISSET(m_fdMother, &m_InputSet))
	{
		cerr << "DEBUG : Accepting new connection." << endl ;

		i = sizeof(insockaddr) ;

		if (getsockname(m_fdMother, (struct sockaddr *) &insockaddr, &i) < 0)
		{
			// error binding socket
			//cerr << "System Error : Error accepting new connection.  Error with getsockname() " << strerror(errno) << endl ;							
			g_Log.Log(CPsyMudLog::logError, "Error accepting new connection with getsockname(). %s", strerror(errno)) ;
			return NULL ;		// change this to throw exception
		}

		if ((newsockfd = accept(m_fdMother, (struct sockaddr *) &insockaddr, &i)) < 0)
		{
			// error binding socket
			//cerr << "System Error : Error accepting new connection.  Error with accept() " << strerror(errno) << endl ;							
			g_Log.Log(CPsyMudLog::logError, "Error accepting new connection with accept(). %s", strerror(errno)) ;
			return NULL ;		// change this to throw exception
		}

		// don't let this puppy block io
		MakeDescriptorNonBlocking(newsockfd) ;

		inf = new CPsyMudConnInfo(newsockfd) ;

		cerr << "DEBUG: Before adding to connection list NeedsToBeDisconnectedDueToQuit = " << inf->NeedsToBeDisconnectedDueToQuit() << endl ;

		m_lstConnections.insert(m_lstConnections.end(), inf) ;

		cerr << " - NeedsToBeDisconnectedDueToQuit = " << inf->NeedsToBeDisconnectedDueToQuit() << endl ;
		cerr << " - newsockfd = " << newsockfd << endl ;
	}
	else
	{
		inf = NULL ;
	}

	return inf ;
}

// works with Linux.  May need code for other platforms.
void CPsyMudComm::MakeDescriptorNonBlocking(int sockfd)
{
	int						iFlags ;

	// make non blocking
	iFlags = fcntl(sockfd, F_GETFL) ;
	iFlags |= O_NONBLOCK ;
	if (fcntl(sockfd, F_SETFL, iFlags) < 0)	
	{
		// error making fd non-blocking
		g_Log.Log(CPsyMudLog::logError, "Error with non blocking setup. %s", strerror(errno)) ;
		// cerr << "System Error : Error with non blocking setup " << strerror(errno) << endl ;									
	}
}


void CPsyMudComm::GoToSleep()
{
	g_Log.Log(CPsyMudLog::logGeneral, "Going to sleep.") ;

	FD_ZERO(&m_InputSet) ;				// zero out the input file descriptor set
	FD_SET(m_fdMother, &m_InputSet) ;		// add the mother socket to the input file descriptor set

		// wait indefinitely (as indicated by NULL time value) to see if we
		// get any input on the next file descriptor (like someone requesting
		// a connection.
	if (select(m_fdMother + 1, &m_InputSet, (fd_set *)0, (fd_set *) 0, NULL) < 0)
	{
			// interrupted with signal
		if (errno == EINTR)	
		{
			g_Log.Log(CPsyMudLog::logGeneral, "Waking up with EINTR to process signal.") ;
		}
		else
		{
			g_Log.Log(CPsyMudLog::logError, "Select coma.") ;
		}
	}
	else
	{
		g_Log.Log(CPsyMudLog::logGeneral, "Waking up with new connection.") ;		
	}
}


void CPsyMudComm::ResetDescriptorSets()
{
		// clear all sets in prep for select()
	FD_ZERO(&m_InputSet) ;
	FD_ZERO(&m_OutputSet) ;
	FD_ZERO(&m_ExceptionSet) ;

		// add mother connection to input set in prep for select()
	FD_SET(m_fdMother, &m_InputSet) ;

		// get the highest descriptor
	int iMaxDescriptor = GetMaxDescriptor() ;
	
	FD_SET(iMaxDescriptor, &m_InputSet) ;
	FD_SET(iMaxDescriptor, &m_OutputSet) ;
	FD_SET(iMaxDescriptor, &m_ExceptionSet) ;
	
}

void CPsyMudComm::PollForActivity()
{
	struct timeval tvNull ;
	tvNull.tv_sec = 0 ;
	tvNull.tv_usec = 0 ;

	if (select(GetMaxDescriptor()+1, &m_InputSet, &m_OutputSet, &m_ExceptionSet, &tvNull) < 0)
	{
		g_Log.Log(CPsyMudLog::logError, "Error in PollForActivity()") ;
	}
}

// iterate through all connections and get the highest file descriptor
// This is required for the select() call to avoid polling descriptors
// that aren't in use.
int CPsyMudComm::GetMaxDescriptor()
{
	int	iMaxDescriptor = m_fdMother ;

	for (vector<CPsyMudConnInfo *>::const_iterator p = m_lstConnections.begin() ; p < m_lstConnections.end() ; p++)
	{
		iMaxDescriptor = (*p)->GetSocketDescriptor() > iMaxDescriptor ? (*p)->GetSocketDescriptor() : iMaxDescriptor ;
	}

	return iMaxDescriptor ;	
}

void CPsyMudComm::RemoveConnection(CPsyMudConnInfo & ci)
{
	g_Log.Log(CPsyMudLog::logTrace, "CPsyMudComm::RemoveConnection") ;


	// TO DO: Change this to use the list::find method
	// (probably requires overloaded operator== on CPsyMudConnInfo)

	int fd  = ci.GetSocketDescriptor() ;

	//list<CPsyMudConnInfo *>::iterator p ;
	
	//p = find(m_lstConnections.begin(), m_lstConnections.end(), ci) ;

	vector<CPsyMudConnInfo *>::iterator p ;

	for (p = m_lstConnections.begin() ; p < m_lstConnections.end() && (*p)->GetSocketDescriptor() != fd ; p++)
		;

	if (p != m_lstConnections.end() && (*p)->GetSocketDescriptor() == fd)
	{			
		(*p)->CloseConnection() ;

		if FD_ISSET(fd, &m_InputSet)
			FD_CLR(fd, &m_InputSet) ;
	
			if FD_ISSET(fd, &m_OutputSet)
			FD_CLR(fd, &m_OutputSet) ;

		if FD_ISSET(fd, &m_ExceptionSet)
			FD_CLR(fd, &m_ExceptionSet) ;

		m_lstConnections.erase(p) ;

	}
	else
	{
		g_Log.Log(CPsyMudLog::logError, "CPsyMudComm::RemoveConnection : Never found the descriptor we wanted to disconnect.") ;
	}

	return ;

}



// ==================================================================================

CPsyMudConnInfo::CPsyMudConnInfo(int sockfd) : m_bufOutputBuffer(DEFAULT_OUTPUT_BUFFER_SIZE, false), m_bufInputBuffer(DEFAULT_INPUT_BUFFER_SIZE, true)
{
	m_fd = sockfd ;
	//m_szHostName = new char[len(szHostName)+1] ;
	//strcpy(m_szHostName, szHostName) ;

	ClearOutputBuffer() ;
	ClearInputBuffer() ;

	// TO DO: Set current time, clear connection states etc.
	
	m_bNeedsToBeDisconnectedDueToCommError = false ;
	m_bNeedsToBeDisconnectedDueToQuit = false ;

	m_bHasPendingCommandInInputBuffer = false ;

	m_bHasPrompt = false ;

	m_bHasPendingCommandInInputBuffer = false ;
}


CPsyMudConnInfo::~CPsyMudConnInfo()
{
	cerr << "DEBUG: Destructor for CPsyMudConnInfo with fd " << m_fd << " called." << endl ;	
}


void CPsyMudConnInfo::CloseConnection()
{
	int iRet ;

	iRet = close(m_fd) ;

	g_Log.Log(CPsyMudLog::logTrace, "CloseConnection() returned %d", iRet) ;
		
}

// to do: Implement some nice output buffer dyamic chaining that way
// we can handle any size output without problems.  Either that or
// go the preallocated-buffer-pool route
void CPsyMudConnInfo::AddToOutputBuffer(const string& sTextToOutput)
{
	string sConverted ;

	m_bufOutputBuffer.AppendToEnd(ConvertColorCodesToANSIColor(sTextToOutput, sConverted)) ;
}

void CPsyMudConnInfo::ProcessOutput(const string& sPrompt)
{
	string sOutput ;
	string sConvertedPrompt ;

	sConvertedPrompt = ConvertColorCodesToANSIColor(sPrompt, sConvertedPrompt) ;

		// this is for when we interrupt them	
	if (m_bHasPrompt)
	{
		sOutput = "\r\n" ;
	}

	m_bufOutputBuffer.AppendToEnd(sConvertedPrompt) ;

	WriteToSocket() ;

	m_bHasPrompt = true ;
}

int CPsyMudConnInfo::WriteToSocket()
{
	int		iBytesLeft ;
	int		iBytesWritten ;

	string 	sTextToWrite = m_bufOutputBuffer.RemoveAll() ;

	iBytesLeft = sTextToWrite.length() ;

	string::iterator p = sTextToWrite.begin() ;

	while (iBytesLeft > 0)
	{
		iBytesWritten = write(m_fd, p, iBytesLeft) ;

		if (iBytesWritten > 0)
		{
			// cool.
			//return BytesWritten ;
			p += iBytesWritten ;
			iBytesLeft -= iBytesWritten ;	
		}
		else if (iBytesWritten == 0)
		{
			// not so cool error.  should never happen
			g_Log.Log(CPsyMudLog::logError, "write() returned zero.  Pretty strange.") ;
			return -1 ;
		}
		else
		{
			if (errno == EAGAIN)
			{
				// write would block and stop the whole mud
				m_bNeedsToBeDisconnectedDueToCommError = true ;

				g_Log.Log(CPsyMudLog::logError, "write() returned EAGAIN.  Socket write would block.  Will disconnect player.") ;

				return -1 ;
			}
			else
			{
				// fatal
				m_bNeedsToBeDisconnectedDueToCommError = true ;

				g_Log.Log(CPsyMudLog::logError, "write() returned fatal error.  Will disconnect player.") ;

				return -1 ;
			}
		}	// if BytesWritten
	}	//while

	m_bufOutputBuffer.Clear() ;

	return 0 ;
}


string CPsyMudConnInfo::ChunkFirstCommandFromInputBuffer()
{
	string sCommandLine = m_bufInputBuffer.RemoveFirstCommandLine() ;

	return sCommandLine ;
}

void CPsyMudConnInfo::ProcessInput()
{
	ReadFromSocket() ;
}

int CPsyMudConnInfo::ReadFromSocket()
{
	int	iBytesRead ;

	static char szTmp[DEFAULT_INPUT_BUFFER_SIZE+1] ;
	memset(szTmp, 0, DEFAULT_INPUT_BUFFER_SIZE+1) ;

	iBytesRead = read(m_fd, szTmp, DEFAULT_INPUT_BUFFER_SIZE) ;

	if (iBytesRead > 0)
	{
		// all is cool.  Null terminate the string and return
		*(szTmp + iBytesRead) = '\0' ;

		m_bufInputBuffer.AppendToEnd(szTmp) ;

		return iBytesRead ;
	}
	else if (iBytesRead == 0)
	{
		// eof: connection broken by the other side
		m_bNeedsToBeDisconnectedDueToCommError = true ;
		g_Log.Log(CPsyMudLog::logGeneral, "Socket read returned EOF (user-side connection broken)") ;		

		return -1 ;
	}
	else
	{
		// interrupted
		if (errno == EINTR)
		{
			return 0 ;
		}
	
		// read would block
		if (errno == EAGAIN)
		{
			// blocking is not a problem on the reads.  we just ignore it and move on
			return 0 ;
		}

		// no idea here
		g_Log.Log(CPsyMudLog::logError, "Socket read returned unknown error.  errno == %d", errno) ;		
		m_bNeedsToBeDisconnectedDueToCommError = true ;

		return -1 ;
	}
}
