//=======================================================================
// This file is a component of PsyMUD written by Peter M. Brown
// copyright            : (C) 2000 by Peter M Brown
// email                : pmbrown@cablespeed.com
//=======================================================================
//                                                                          
// This program is free software; you can redistribute it and/or modify  
// it under the terms of the GNU General Public License as published by  
// the Free Software Foundation; either version 2 of the License, or     
// (at your option) any later version. Pleae keep all headers and 
// documentation (including development credits) intact.
//                                                                          
//=======================================================================


#include "CPsyMudUniverse.h"
#include "CPsyMudRoomExit.h"
#include "CPsyMudItem.h"
#include <fstream>
#include <stdlib.h>		// for atoi
#include <dirent.h>		// for directory reading
#include "fileutil.h"	// for file manipulation functions
#include "stringutil.h"

#define IsComment(s)	(s[0] == '#' || (s[0] == '/' && s[1] == '/'))

#define AddRoom(rm)				m_Rooms.insert(make_pair(rm->GetObjectID(), rm))
#define AddItemTemplate(itm)	m_ItemTemplates.insert(make_pair(itm->GetObjectID(), itm))


CPsyMudUniverse::CPsyMudUniverse()
{
	cerr << "(DEBUG) : CPsyMudUniverse constructor called." << endl ;
}

void CPsyMudUniverse::SetDirectory (const char * szDirectory)
{
	m_sUniverseDirectory = szDirectory ;
}

void CPsyMudUniverse::LoadUniverse() throw(CPsyMudUniverseException)
{
	DIR * 				dir ;
	struct dirent * 	ent ;
	string				sFileName ;

	// ===================================================================================
	// first, we have to load all the items ==============================================

	if (!(dir = opendir(m_sUniverseDirectory.c_str())))
	{
		cerr << "Error : Could not open universe directory" << endl ;
		exit (1) ;
	}
	while ((ent = readdir(dir)))
	{
		if ((string(ent->d_name) != ".") && (string(ent->d_name) != ".."))
		{
			if (ExtensionFromFileName(ent->d_name) == "item")
			{
				LoadItems(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
		}	
	}

	closedir(dir) ;


	// ===================================================================================
	// second, we have to load all the NPCs ==============================================

	if (!(dir = opendir(m_sUniverseDirectory.c_str())))
	{
		cerr << "Error : Could not open universe directory" << endl ;
		exit (1) ;
	}
	while ((ent = readdir(dir)))
	{
		if ((string(ent->d_name) != ".") && (string(ent->d_name) != ".."))
		{
			if (ExtensionFromFileName(ent->d_name) == "npc")
			{
				LoadNPCs(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
		}	
	}

	closedir(dir) ;

	// ===================================================================================
	// load the rest of the stuff in the dir =============================================


	if (!(dir = opendir(m_sUniverseDirectory.c_str())))
	{
		cerr << "Error : Could not open universe directory" << endl ;
		exit (1) ;
	}

	while ((ent = readdir(dir)))
	{
		if ((string(ent->d_name) != ".") && (string(ent->d_name) != ".."))
		{
			if (ExtensionFromFileName(ent->d_name) == "world")
			{
				LoadWorlds(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
			else if (ExtensionFromFileName(ent->d_name) == "zone")
			{
				LoadZones(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
			else if (ExtensionFromFileName(ent->d_name) == "room")
			{
				LoadRooms(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
			else if (ExtensionFromFileName(ent->d_name) == "script")
			{
				LoadScripts(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
			else if (ExtensionFromFileName(ent->d_name) == "login")
			{
				LoadLoginBanner(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
			else if (ExtensionFromFileName(ent->d_name) == "logout")
			{
				LoadLogoutBanner(m_sUniverseDirectory + "/" + string(ent->d_name)) ;
			}
		}	
	}

	closedir(dir) ;

	ResolveIDsIntoPointers() ;
}

CPsyMudRoom * CPsyMudUniverse::GetRoomByNumber(unsigned long ulRoomNumber)
{
	map<unsigned long, CPsyMudRoom *>::const_iterator p ;

	p = m_Rooms.find(ulRoomNumber) ;

	if (p != m_Rooms.end())
	{
		// cerr << "DEBUG: Room " << ulRoomNumber << " WAS found." << endl ;
		return (*p).second ;
	}
	else
	{
		cerr << "DEBUG: Room " << ulRoomNumber << " was not found." << endl ;
		return NULL ;
	}
}


CPsyMudItem * CPsyMudUniverse::GetItemTemplateByNumber(unsigned long ulItemNumber)
{
	map<unsigned long, CPsyMudItem *>::const_iterator p ;

	p = m_ItemTemplates.find(ulItemNumber) ;

	if (p != m_ItemTemplates.end())
	{
		return (*p).second ;
	}
	else
	{
		cerr << "DEBUG: Item " << ulItemNumber << " was not found." << endl ;
		return NULL ;
	}
}

CPsyMudItem * CPsyMudUniverse::LoadItemIntoRoom(unsigned long ulItemNumber, unsigned long ulRoomNumber)
{
	CPsyMudItem * pNewItem ;
	CPsyMudItem * pTemplateItem ;
	CPsyMudRoom * pLoadRoom ;

	if (pTemplateItem = GetItemTemplateByNumber(ulItemNumber))
	{
		if (pLoadRoom = GetRoomByNumber(ulRoomNumber))
		{
			// construct new item based on template item
			pNewItem = new CPsyMudItem(*pTemplateItem) ;

			// add item to room
			pLoadRoom->GetItems()->AddItem(pNewItem) ;

			return pNewItem ;
		}
		else
		{
			// failed because room number was bad	
			return NULL ;
		}
	}
	else
	{	
		// failed because item does not exist
		return NULL ;
	}
}


void CPsyMudUniverse::ResolveIDsIntoPointers()
{
	CPsyMudRoom * pCurrentRoom ;
	CPsyMudRoom * pDestination ;

	g_Log.Log(CPsyMudLog::logTrace, "Resolving IDs in exits into object pointers...") ;

	for (map<unsigned long, CPsyMudRoom *>::iterator pRoom = m_Rooms.begin() ; pRoom != m_Rooms.end() ; pRoom++)
	{
		pCurrentRoom = pRoom->second ;

		// loop through all exits in the room
		for (vector<CPsyMudRoomExit *>::iterator pExit = pCurrentRoom->GetExitList()->begin() ; pExit != pCurrentRoom->GetExitList()->end() ; pExit++)
		{
			pDestination = GetRoomByNumber((*pExit)->GetTemporaryDest()) ;

			if (pDestination)
			{
				(*pExit)->SetDestination(pDestination) ;
			}
			else
			{
				g_Log.Log(CPsyMudLog::logError, "Unresolved room reference \"%u\"in exit \"%s\" of room \"%u\"", (*pExit)->GetTemporaryDest(), (*pExit)->GetDirectionName().c_str(), pCurrentRoom->GetObjectID()) ;
			}
		}

	}

	g_Log.Log(CPsyMudLog::logTrace, "Finished resolving IDs in exits.") ;
}


void CPsyMudUniverse::LoadLoginBanner(const string sFileName)
{
	ifstream	ifs ;
	char		buf[512] ;

	// get login banner ----------------------------

	ifs.open(sFileName.c_str(), ios::in) ;

	if (!ifs)
	{
		//input failed
	}
	else
	{
		while (!ifs.eof())
		{
			ifs.getline(buf, sizeof(buf), '\n') ;
			if (!(ifs.eof() && strlen(buf) == 0))
			{
				m_sLoginBanner = m_sLoginBanner + buf + "\n\r";
			}
		}
	}

	ifs.close() ;
}

void CPsyMudUniverse::LoadLogoutBanner(const string sFileName)
{
	ifstream	ifs ;
	char		buf[512] ;

	// get logout banner ----------------------------

	ifs.open(sFileName.c_str(), ios::in) ;

	if (!ifs)
	{
		//input failed
	}
	else
	{
		while (!ifs.eof())
		{
			ifs.getline(buf, sizeof(buf), '\n') ;
			if (!(ifs.eof() && strlen(buf) == 0))
			{
				m_sLogoutBanner = m_sLogoutBanner + buf + "\n\r";
			}
		}
	}

	ifs.close() ;
}

void CPsyMudUniverse::LoadWorlds(const string sFileName)
{

}

void CPsyMudUniverse::LoadZones(const string sFileName)
{
}


void CPsyMudUniverse::LoadRooms(const string sFileName)
{
	ifstream	ifs ;
	char		buf[512] ;
	string		s ;
	string		sRoomFileData ;
	string		sSingleRoomData ;

	g_Log.Log(CPsyMudLog::logGeneral, "Loading rooms from file %s", sFileName.c_str()) ;

	ifs.open(sFileName.c_str(), ios::in) ;

	if (!ifs)
	{
		//input failed
	}
	else
	{
		sRoomFileData = "" ;
		sRoomFileData.reserve(10000) ;	// might need to change this to some other later

		// reads whole file into memory.  Hope they keep the files small :-)
		while (!ifs.eof())
		{
			ifs.getline(buf, sizeof(buf), '\n') ;
			
			s = StripTrailingBlanks(StripLeadingBlanks(string(buf))) ;

			s+='\n' ;

			if (!IsComment(s))
			{
				sRoomFileData += s ;
				memset(buf, sizeof(buf), 0) ;
			}
		}

		ifs.close() ;

		sSingleRoomData = RemoveEverythingBetweenTags(sRoomFileData, "<room>", "</room>") ;

		CPsyMudRoom * rm ;

		while (sSingleRoomData != "")
		{		

			// create new room object
			rm = new CPsyMudRoom ;			


			string sProperty ;
			string sSingleExitData ;

			sProperty = RemoveEverythingBetweenTags(sSingleRoomData, "<name>", "</name>") ;
			rm->SetName(sProperty) ;

			sProperty = RemoveEverythingBetweenTags(sSingleRoomData, "<zone>", "</zone>") ;

			sProperty = RemoveEverythingBetweenTags(sSingleRoomData, "<number>", "</number>") ;
			rm->SetObjectID(atol(sProperty.c_str())) ;


			sProperty = RemoveEverythingBetweenTags(sSingleRoomData, "<description>", "</description>") ;
			rm->SetDescription(sProperty) ;

			// add the room to the room list
			//m_Rooms.insert(m_Rooms.end(), rm) ;
			AddRoom(rm) ;



						
			// load all room items			
			string 	sAllItems ;
			sAllItems = RemoveEverythingBetweenTags(sSingleRoomData, "<items>", "</items>") ;

			if (sAllItems != "")
			{
				CPsyMudItem * itm ;
				string 	sSingleItem ;
				sSingleItem = RemoveEverythingBetweenTags(sAllItems, "<item>", "</item>") ;

				while (sSingleItem != "")
				{
					if (itm = LoadItemIntoRoom(atol(sSingleItem.c_str()), rm->GetObjectID()))
					{
						cerr << "DEBUG: Load object " << sSingleItem << " into room " << rm->GetObjectID() << endl ;
					}
					else
					{
						g_Log.Log(CPsyMudLog::logError, "Could not load object %u into room %u",atol(sSingleItem.c_str()), rm->GetObjectID()) ;
					}

					sSingleItem = RemoveEverythingBetweenTags(sAllItems, "<item>", "</item>") ;
				}
			}


			// Get exit information from room and add room exits as well.
			CPsyMudRoomExit * 	ex ;
			sSingleExitData = RemoveEverythingBetweenTags(sSingleRoomData, "<exit>", "</exit>") ;

			while (sSingleExitData != "")
			{
				ex = new CPsyMudRoomExit ;	

				//cerr << "DEBUG: sSingleExitData = " << sSingleExitData << endl ;

				sProperty = RemoveEverythingBetweenTags(sSingleExitData, "<name>", "</name>") ;
				ex->SetDirectionName(sProperty) ;

				sProperty = RemoveEverythingBetweenTags(sSingleExitData, "<description>", "</description>") ;
				ex->SetDescription(sProperty) ;

				sProperty = RemoveEverythingBetweenTags(sSingleExitData, "<destination>", "</destination>") ;
				ex->SetTemporaryDest(atoi(sProperty.c_str())) ;

				sProperty = RemoveEverythingBetweenTags(sSingleExitData, "<defaultlocked>", "</defaultlocked>") ;
				sProperty = RemoveEverythingBetweenTags(sSingleExitData, "<defaultclosed>", "</defaultclosed>") ;

				if (!(rm->AddExit(ex)))
				{
					g_Log.Log(CPsyMudLog::logError, "Error adding exit \"%s\" to room %ud", ex->GetDirectionName().c_str(), rm->GetObjectID()) ;
				}

				sSingleExitData = RemoveEverythingBetweenTags(sSingleRoomData, "<exit>", "</exit>") ;
			}



			// get next room from file - if there is another room
			sSingleRoomData = RemoveEverythingBetweenTags(sRoomFileData, "<room>", "</room>") ;
		}
	}

	g_Log.Log(CPsyMudLog::logGeneral, "Finished Loading rooms from file %s", sFileName.c_str()) ;

}

void CPsyMudUniverse::LoadNPCs(const string sFileName)
{
}

void CPsyMudUniverse::LoadItems(const string sFileName)
{
	ifstream	ifs ;
	char		buf[512] ;
	string		s ;
	string		sItemFileData ;
	string		sSingleItemData ;

	unsigned long int	lNumItemsLoaded = 0 ;

	g_Log.Log(CPsyMudLog::logGeneral, "Loading items from file %s", sFileName.c_str()) ;

	ifs.open(sFileName.c_str(), ios::in) ;

	if (!ifs)
	{
		//input failed
	}
	else
	{
		sItemFileData = "" ;
		sItemFileData.reserve(10000) ;	// might need to change this to some other later

		// reads whole file into memory.  Hope they keep the files small :-)
		while (!ifs.eof())
		{
			ifs.getline(buf, sizeof(buf), '\n') ;
			
			s = StripTrailingBlanks(StripLeadingBlanks(string(buf))) ;

			s+='\n' ;

			if (!IsComment(s))
			{
				sItemFileData += s ;
				memset(buf, sizeof(buf), 0) ;
			}
		}

		ifs.close() ;

		sSingleItemData = RemoveEverythingBetweenTags(sItemFileData, "<item>", "</item>") ;

		CPsyMudItem * itm ;

		while (sSingleItemData != "")
		{		

			// create new room object
			itm = new CPsyMudItem ;			
			lNumItemsLoaded++ ;

			string sProperty ;

			sProperty = RemoveEverythingBetweenTags(sSingleItemData, "<number>", "</number>") ;
			itm->SetObjectID(atol(sProperty.c_str())) ;

			if (itm->GetObjectID() != 0)
			{
				string sAllAliases ;

				// add aliases.  Aliases are words which can be used to refer to an object
				sAllAliases = RemoveEverythingBetweenTags(sSingleItemData, "<aliases>", "</aliases>") ;
				sProperty = RemoveEverythingBetweenTags(sAllAliases, "<alias>", "</alias>") ;
				while (sProperty != "")
				{
					itm->AddAlias(sProperty) ;
					cerr << "DEBUG: Object Alias Added : " << sProperty << " for object " << itm->GetObjectID() << endl ;

					sProperty = RemoveEverythingBetweenTags(sAllAliases, "<alias>", "</alias>") ;

				}


				sProperty = RemoveEverythingBetweenTags(sSingleItemData, "<shortdescription>", "</shortdescription>") ;
				itm->SetShortDescription(sProperty) ;

				sProperty = RemoveEverythingBetweenTags(sSingleItemData, "<longdescription>", "</longdescription>") ;
				itm->SetLongDescription(sProperty) ;

				// Get names for this item


				AddItemTemplate(itm) ;
			}
			else
			{
				g_Log.Log(CPsyMudLog::logError, "Error in item file %s.  Item with zero ID found.", sFileName.c_str()) ;
			}

			// get next room from file - if there is another room
			sSingleItemData = RemoveEverythingBetweenTags(sItemFileData, "<item>", "</item>") ;
		}
	}

	g_Log.Log(CPsyMudLog::logGeneral, "Finished Loading items from file %s.  %d items were loaded", sFileName.c_str(), lNumItemsLoaded) ;

}

void CPsyMudUniverse::LoadScripts(const string sFileName)
{
}
	
